use std::{cell::RefCell, collections::HashMap, rc::Rc};

use crate::{
    DeBrujinAst,
    inference::{
        recursive::infer_type as infer_type_rec, unification::infer_type as infer_type_uni,
    },
    multi_map::MultiMap,
    parse::{Ast, Constant},
    types::{PrimitiveType, TaggedType, Type, TypeTag},
};

use super::unification::TypeVarCtx;

#[test]
fn infer_add_nat_rec() {
    let ast: DeBrujinAst = Ast::Application(
        Box::new(Ast::Application(
            Box::new(Ast::Variable("add".to_string())),
            Box::new(Ast::Constant(Constant::Nat(5))),
        )),
        Box::new(Ast::Constant(Constant::Nat(7))),
    )
    .into();

    let mut gamma = HashMap::new();
    gamma.insert(
        "add".to_string(),
        crate::types::TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::Arrow(
                Box::new(Type::Variable("a".to_string())),
                Box::new(Type::Arrow(
                    Box::new(Type::Variable("a".to_string())),
                    Box::new(Type::Variable("a".to_string())),
                )),
            ))),
        ),
    );

    let infered = infer_type_rec(&gamma, ast).unwrap();
    assert_eq!(
        infered,
        TaggedType::Concrete(Type::Primitive(PrimitiveType::Nat))
    );
}

#[test]
fn infer_add_nat_partial_rec() {
    let ast: DeBrujinAst = Ast::Application(
        Box::new(Ast::Variable("add".to_string())),
        Box::new(Ast::Constant(Constant::Nat(5))),
    )
    .into();

    let mut gamma = HashMap::new();
    gamma.insert(
        "add".to_string(),
        crate::types::TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::Arrow(
                Box::new(Type::Variable("a".to_string())),
                Box::new(Type::Arrow(
                    Box::new(Type::Variable("a".to_string())),
                    Box::new(Type::Variable("a".to_string())),
                )),
            ))),
        ),
    );

    let infered = infer_type_rec(&gamma, ast).unwrap();
    assert_eq!(
        infered,
        TaggedType::Concrete(Type::arrow(PrimitiveType::Nat, PrimitiveType::Nat))
    );
}

#[test]
fn subtype_constraints() {
    let typ = TaggedType::Tagged(
        TypeTag::Num,
        "a".to_string(),
        Box::new(Type::arrow("a", "a").into()),
    );

    let st_constraints = Rc::new(RefCell::new(MultiMap::new()));
    let ctx = TypeVarCtx::new();

    let res = typ.make_constraints(st_constraints.clone(), &ctx);

    assert_eq!(
        st_constraints
            .borrow()
            .get(&Type::Variable("?typ_0".to_string())),
        vec![&TypeTag::Num]
    );

    assert_eq!(res, Type::arrow("?typ_0", "?typ_0"));
}

#[test]
fn infer_add_nat_uni() {
    let ast: DeBrujinAst = Ast::Application(
        Box::new(Ast::Application(
            Box::new(Ast::Variable("add".to_string())),
            Box::new(Ast::Constant(Constant::Nat(5))),
        )),
        Box::new(Ast::Constant(Constant::Nat(7))),
    )
    .into();

    let mut gamma = HashMap::new();
    gamma.insert(
        "add".to_string(),
        TaggedType::Tagged(
            TypeTag::Num,
            "a".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow(
                "a",
                Type::arrow("a", "a"),
            ))),
        ),
    );

    let infered = infer_type_uni(&gamma, ast).unwrap();
    assert_eq!(
        infered,
        TaggedType::Concrete(Type::Primitive(PrimitiveType::Nat))
    );
}

#[test]
fn infer_add_nat_partial_uni() {
    let ast: DeBrujinAst = Ast::Application(
        Box::new(Ast::Variable("add".to_string())),
        Box::new(Ast::Constant(Constant::Nat(5))),
    )
    .into();

    let mut gamma = HashMap::new();
    gamma.insert(
        "add".to_string(),
        TaggedType::Concrete(Type::arrow(
            PrimitiveType::Nat,
            Type::arrow(PrimitiveType::Nat, PrimitiveType::Nat),
        )),
    );

    let typ = infer_type_uni(&gamma, ast).unwrap();

    assert_eq!(
        typ,
        TaggedType::Concrete(Type::arrow(PrimitiveType::Nat, PrimitiveType::Nat))
    );
}

#[test]
fn infer_id_uni() {
    let ast: DeBrujinAst = Ast::Abstraction(
        "x".to_string(),
        None,
        Box::new(Ast::Variable("x".to_string())),
    )
    .into();

    let gamma = HashMap::new();
    let typ = infer_type_uni(&gamma, ast).unwrap();

    assert_eq!(
        typ,
        TaggedType::Tagged(
            TypeTag::Any,
            "?typ_0".to_string(),
            Box::new(TaggedType::Concrete(Type::arrow("?typ_0", "?typ_0")))
        )
    );
}

#[test]
fn subst_type_var() {
    let typ = Type::arrow("a", "?typ_4");

    let subst = Type::arrow("b", "b");

    let typ = typ.subst_typevar("?typ_4", &subst);

    assert_eq!(typ, (Type::arrow("a", Type::arrow("b", "b"))));
}
